@@ -3,19 +3,22 @@
  *
  * @APPLE_LICENSE_HEADER_START@
  * 
- * The contents of this file constitute Original Code as defined in and
- * are subject to the Apple Public Source License Version 1.1 (the
- * "License").  You may not use this file except in compliance with the
- * License.  Please obtain a copy of the License at
- * http://www.apple.com/publicsource and read it before using this file.
+ * Copyright (c) 1999-2003 Apple Computer, Inc.  All Rights Reserved.
  * 
- * This Original Code and all software distributed under the License are
- * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
+ * This file contains Original Code and/or Modifications of Original Code
+ * as defined in and that are subject to the Apple Public Source License
+ * Version 2.0 (the 'License'). You may not use this file except in
+ * compliance with the License. Please obtain a copy of the License at
+ * http://www.opensource.apple.com/apsl/ and read it before using this
+ * file.
+ * 
+ * The Original Code and all software distributed under the License are
+ * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
  * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
  * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
- * License for the specific language governing rights and limitations
- * under the License.
+ * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
+ * Please see the License for the specific language governing rights and
+ * limitations under the License.
  * 
  * @APPLE_LICENSE_HEADER_END@
  */
@@ -30,18 +33,22 @@
 #ifndef _IOKIT_APPLEMACRISCPCI_H
 #define _IOKIT_APPLEMACRISCPCI_H
 
+#include <IOKit/IOBufferMemoryDescriptor.h>
 #include <IOKit/pci/IOPCIBridge.h>
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
 enum {
-    kBridgeSelfDevice = 11
+    kHTBridgeSelfDevice = 0,
+    kPCIBridgeSelfDevice = 11
 };
 
 enum {
-  kMacRISCAddressSelect 	= 0x48,
-  kMacRISCModeSelect		= 0x50,
-  kMacRISCModeSelectRDGBit 	= 0x00080000
+  kMacRISCHTAddressSelect 	= 0x80,
+  kMacRISCPCIAddressSelect 	= 0x48,
+  kMacRISCPCIModeSelect		= 0x50,
+  kMacRISCPCIModeSelectRDGBit 	= 0x00080000,
+  kMacRISCPCIModeSelectWCBit    = 0x00100000
 };
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -61,7 +68,8 @@ protected:
 
     UInt16			coarseAddressMask;
     UInt16			fineAddressMask;
-    UInt8			primaryBus;
+    UInt8			secondaryBus;
+    UInt8			subordinateBus;
     UInt8			configDataOffsetMask;
     UInt32			uniNVersion;
 
@@ -101,6 +109,47 @@ public:
 
 };
 
+class AppleMacRiscHT : public IOPCIBridge
+{
+    OSDeclareDefaultStructors(AppleMacRiscHT)
+
+protected:
+    IOMemoryMap *		configType0Map;
+    IOMemoryMap *		configType1Map;
+    IOMemoryMap *		configSelfMap;
+    IODeviceMemory *		ioMemory;
+
+    volatile UInt8	*	configType0;
+    volatile UInt8	*	configType1;
+    volatile UInt8	*	configSelf;
+
+    UInt16			coarseAddressMask;
+    UInt16			fineAddressMask;
+    UInt8			primaryBus;
+
+    inline volatile UInt8 * setConfigSpace( IOPCIAddressSpace space, UInt8 offset );
+
+public:
+    virtual bool start(	IOService * provider );
+    virtual bool configure( IOService * provider );
+
+    virtual void free();
+
+    virtual IODeviceMemory * ioDeviceMemory( void );
+
+    virtual UInt32 configRead32( IOPCIAddressSpace space, UInt8 offset );
+    virtual void configWrite32( IOPCIAddressSpace space,
+					UInt8 offset, UInt32 data );
+    virtual UInt16 configRead16( IOPCIAddressSpace space, UInt8 offset );
+    virtual void configWrite16( IOPCIAddressSpace space,
+					UInt8 offset, UInt16 data );
+    virtual UInt8 configRead8( IOPCIAddressSpace space, UInt8 offset );
+    virtual void configWrite8( IOPCIAddressSpace space,
+					UInt8 offset, UInt8 data );
+
+    virtual IOPCIAddressSpace getBridgeSpace( void );
+};
+
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
 /* Definitions of UniNorth Target config registers */
@@ -108,13 +157,23 @@ enum {
     kUniNGART_BASE		= 0x8c,
     kUniNAGP_BASE		= 0x90,
     kUniNGART_CTRL		= 0x94,
-    kUniNINTERNAL_STATUS	= 0x98
+    kUniNINTERNAL_STATUS	= 0x98,
+    kUniNDUMMY_PAGE		= 0xa4
 };
 enum {
     kGART_INV			= 0x00000001,
     kGART_EN			= 0x00000100,
     kGART_2xRESET		= 0x00010000,
-    kGART_DISSBADET		= 0x00020000
+    kGART_DIS_SBA_DET		= 0x00020000,
+    kGART_SYNC_MODE		= 0x00040000,
+    kGART_PERF_RD		= 0x00080000
+};
+
+
+enum {
+    kIOAGP3Mode			= 0x00000008,
+    kIOAGP4xDataRateMode3	= 0x00000001,
+    kIOAGP8xDataRateMode3	= 0x00000002,
 };
 
 class IORangeAllocator;
@@ -134,7 +193,11 @@ protected:
     IOByteCount		gartLength;
     UInt32		gartCtrl;
     UInt32		agpCommandMask;
+    UInt32		agpCommandSet;
     UInt8		targetAGPRegisters;
+    UInt8		isU3;
+    UInt8		isU32;
+    IOBufferMemoryDescriptor * dummyPage;
 
 private:
     virtual IOReturn setAGPEnable( IOAGPDevice * master, bool enable,
